# :title: Text::Format
# :main: Text::Format
#--
# Text::Format for Ruby
# Version 0.52.2
#
# Copyright (c) 2002 Austin Ziegler
#
# $Id: $
#
# ==========================================================================
# Revision History ::
# YYYY.MM.DD    Change ID       Developer
#               Description
# --------------------------------------------------------------------------
# 2002.10.18                    Austin Ziegler
#               Fixed a minor problem with tabs not being counted. Changed
#               abbreviations from Hash to Array to better suit Ruby's
#               capabilities. Fixed problems with the way that Array arguments
#               are handled in calls to the major object types, excepting in
#               Text::Format#expand and Text::Format#unexpand (these will
#               probably need to be fixed).
# 2002.10.30                    Austin Ziegler
#               Fixed the ordering of the <=> for binary tests. Fixed
#               Text::Format#expand and Text::Format#unexpand to handle array
#               arguments better.
#
# ==========================================================================
#++
module Text #:nodoc:
        # = Introduction
        #
        # Text::Format provides the ability to nicely format fixed-width text
        # with knowledge of the writeable space (number of columns), margins,
        # and indentation settings.
        #
        # Copyright::   Copyright (c) 2002 by Austin Ziegler
        # Version::     0.52.2
        # Based On::    Perl
        #               Text::Format[http://search.cpan.org/author/GABOR/Text-Format0.52/lib/Text/Format.pm],
        #               Copyright (c) 1998 Gbor Egressy
        # Licence::     Ruby's, Perl Artistic, or GPL version 2 (or later)
        #
    class Format
        VERSION = '0.52.2'

            # Local abbreviations. More can be added with
            # Text::Format.abbreviations
        ABBREV = [ 'Mr', 'Mrs', 'Ms', 'Jr', 'Sr' ]

            # Formatting values
        LEFT_ALIGN  = 0
        RIGHT_ALIGN = 1
        RIGHT_FILL  = 2
        JUSTIFY     = 3


        include Comparable

    private
        LEQ_RE = /[.?!]['"]?$/

        def brk_re(i)
            /((?:\S+\s+){#{i}})(.+)/
        end

        def posint(p) #:nodoc:
            p.to_i.abs
        end

    public
        def <=>(o)
            t = @text <=> o.text
            return t unless t == 0
            t = @columns <=> o.columns
            return t unless t == 0
            t = @left_margin <=> o.left_margin
            return t unless t == 0
            t = @right_margin <=> o.right_margin
            return t unless t == 0
            t = @first_indent <=> o.first_indent
            return t unless t == 0
            t = @body_indent <=> o.body_indent
            return t unless t == 0
            t = @tag_text <=> o.tag_text
            return t unless t == 0
            t = @tabstop <=> o.tabstop
            return t unless t == 0
            t = @format_style <=> o.format_style
            return t unless t == 0
            if @extra_space != o.extra_space
                return (@extra_space ? 1 : -1)
            end
            if @tag_paragraph != o.tag_paragraph
                return (@tag_paragraph ? 1 : -1)
            end
            if @nobreak != o.nobreak
                return (@nobreak ? 1 : -1)
            end
            if @abbreviations.empty? || o.abbreviations.empty?
                if @abbreviations.empty? && !o.abbreviations.empty?
                    t = -1
                elsif !@abbreviations.empty? && o.abbreviations.empty?
                    t = 1
                else
                    t = 0
                end
            end
            return t unless t == 0
            if @nobreak_regex.empty? || o.nobreak_regex.empty?
                if @nobreak_regex.empty? && !o.nobreak_regex.empty?
                    t = -1
                elsif !@nobreak_regex.empty? && o.nobreak_regex.empty?
                    t = 1
                else
                    t = 0
                end
            end
            t
        end

            # The text to be manipulated. Note that value is optional, but if
            # the formatting functions are called without values, this text
            # is what will be formatted.
            #
            # *Default*::       <tt>[]</tt>
            # <b>Used in</b>::  All methods
        attr_accessor :text
            # The total width of the format area. The margins, indentation,
            # and text are formatted into this space.
            #
            #                             COLUMNS
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   left margin  indent  text is formatted into here  right margin
            #
            # *Default*::       <tt>72</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>,
            #                   <tt>#center</tt>
        attr_reader :columns

            # The total width of the format area. The margins, indentation,
            # and text are formatted into this space. The value provided is
            # silently converted to a positive integer.
            #
            #                             COLUMNS
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   left margin  indent  text is formatted into here  right margin
            #
            # *Default*::       <tt>72</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>,
            #                   <tt>#center</tt>
        def columns=(c)
            @columns = posint(c)
        end

            # The number of spaces used for the left margin.
            #
            #                             columns
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   LEFT MARGIN  indent  text is formatted into here  right margin
            #
            # *Default*::       <tt>0</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>,
            #                   <tt>#center</tt>
        attr_reader :left_margin

            # The number of spaces used for the left margin. The value
            # provided is silently converted to a positive integer value.
            #
            #                             columns
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   LEFT MARGIN  indent  text is formatted into here  right margin
            #
            # *Default*::       <tt>0</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>,
            #                   <tt>#center</tt>
        def left_margin=(l)
            @left_margin = posint(l)
        end

            # The number of spaces used for the right margin.
            #
            #                             columns
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   left margin  indent  text is formatted into here  RIGHT MARGIN
            #
            # *Default*::       <tt>0</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>,
            #                   <tt>#center</tt>
        attr_reader :right_margin

            # The number of spaces used for the right margin. The value
            # provided is silently converted to a positive integer value.
            #
            #                             columns
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   left margin  indent  text is formatted into here  RIGHT MARGIN
            #
            # *Default*::       <tt>0</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>,
            #                   <tt>#center</tt>
        def right_margin=(r)
            @right_margin = posint(r)
        end

            # The number of spaces to indent the first line of a paragraph.
            #
            #                             columns
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   left margin  INDENT  text is formatted into here  right margin
            #
            # *Default*::       <tt>4</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        attr_reader :first_indent

            # The number of spaces to indent the first line of a paragraph.
            # The value provided is silently converted to a positive integer
            # value.
            #
            #                             columns
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   left margin  INDENT  text is formatted into here  right margin
            #
            # *Default*::       <tt>4</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        def first_indent=(f)
            @first_indent = posint(f)
        end

            # The number of spaces to indent all lines after the first line of
            # a paragraph.
            #
            #                             columns
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   left margin  INDENT  text is formatted into here  right margin
            #
            # *Default*::       <tt>0</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        attr_reader :body_indent

            # The number of spaces to indent all lines after the first line of
            # a paragraph. The value provided is silently converted to a
            # positive integer value.
            #
            #                             columns
            #  <-------------------------------------------------------------->
            #  <-----------><------><---------------------------><------------>
            #   left margin  INDENT  text is formatted into here  right margin
            #
            # *Default*::       <tt>0</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        def body_indent=(b)
            @body_indent = posint(b)
        end

            # Indicates whether sentence terminators should be followed by a
            # single space (+false+), or two spaces (+true+).
            #
            # *Default*::       +false+
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        attr_accessor :extra_space

            # Defines the current abbreviations as an array. This is only used
            # if extra_space is turned on.
            #
            # If one is abbreviating "President" as "Pres." (abbreviations =
            # ["Pres"]), then the results of formatting will be as
            # illustrated in the table below
            #
            #       extra_space  |  include?        |  !include?
            #         true       |  Pres. Lincoln   |  Pres.  Lincoln
            #         false      |  Pres. Lincoln   |  Pres. Lincoln
            #
            # *Default*::       <tt>{}</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        attr_accessor :abbreviations

            # Indicates whether the formatting of paragraphs should be done
            # with tagged paragraphs. Useful only with <tt>#tag_text</tt>.
            #
            # *Default*::       +false+
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        attr_accessor :tag_paragraph

            # The array of text to be placed before each paragraph when
            # <tt>#tag_paragraph</tt> is +true+. When <tt>#format()</tt> is
            # called, only the first element of the array is used. When
            # <tt>#paragraphs</tt> is called, then each entry in the array
            # will be used once, with corresponding paragraphs. If the tag
            # elements are exhausted before the text is exhausted, then the
            # remaining paragraphs will not be tagged. Regardless of
            # indentation settings, a blank line will be inserted between all
            # paragraphs when <tt>#tag_paragraph</tt> is +true+.
            #
            # *Default*::       <tt>[]</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        attr_accessor :tag_text

            # Indicates whether or not the non-breaking space feature should
            # be used
            #
            # *Default*::       +false+
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        attr_accessor :nobreak

            # A hash which holds the regular expressions on which spaces
            # should not be broken. The hash is set up such that the key is
            # the first word and the value is the second word.
            #
            # For example, if +nobreak_regex+ contains the following hash:
            #
            #   { '^Mrs?\.$' => '\S+$', '^\S+$' => '^(?:S|J)r\.$'}
            #
            # Then "Mr. Jones", "Mrs. Jones", and "Jones Jr." would not be
            # broken. If this simple matching algorithm indicates that there
            # should not be a break at the current end of line, then a
            # backtrack is done until there are two words on which line
            # breaking is permitted. If two such words are not found, then the
            # end of the line will be broken *regardless*. If there is a
            # single word on the current line, then no backtrack is done and
            # the word is stuck on the end.
            #
            # *Default*::       <tt>{}</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        attr_accessor :nobreak_regex

            # Indicates the number of spaces that a single tab represents.
            #
            # *Default*::       <tt>8</tt>
            # <b>Used in</b>::  <tt>#expand</tt>, <tt>#unexpand</tt>,
            #                   <tt>#paragraphs</tt>
        attr_reader :tabstop

            # Indicates the number of spaces that a single tab represents.
            #
            # *Default*::       <tt>8</tt>
            # <b>Used in</b>::  <tt>#expand</tt>, <tt>#unexpand</tt>,
            #                   <tt>#paragraphs</tt>
        def tabstop=(t)
            @tabstop = posint(t)
        end

            # Specifies the format style. Allowable values are:
            # [+LEFT_ALIGN+]    Left justified, ragged right.
            #      |A paragraph that is|
            #      |left aligned.|
            # [+RIGHT_ALIGN+]   Right justified, ragged left.
            #      |A paragraph that is|
            #      |     right aligned.|
            # [+RIGHT_FILL+]    Left justified, right ragged, filled to
            #                   width by spaces. (Essentially the same as
            #                   +LEFT_ALIGN+ except that lines are padded
            #                   on the right.)
            #      |A paragraph that is|
            #      |left aligned.      |
            # [+JUSTIFY+]       Fully justified, words filled to width by
            #                   spaces, except the last line.
            #      |A paragraph  that|
            #      |is     justified.|
            #
            # *Default*::       <tt>Text::Format::LEFT_ALIGN</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        attr_reader :format_style

            # Specifies the format style. Allowable values are:
            # [+LEFT_ALIGN+]    Left justified, ragged right.
            #      |A paragraph that is|
            #      |left aligned.|
            # [+RIGHT_ALIGN+]   Right justified, ragged left.
            #      |A paragraph that is|
            #      |     right aligned.|
            # [+RIGHT_FILL+]    Left justified, right ragged, filled to
            #                   width by spaces. (Essentially the same as
            #                   +LEFT_ALIGN+ except that lines are padded
            #                   on the right.)
            #      |A paragraph that is|
            #      |left aligned.      |
            # [+JUSTIFY+]       Fully justified, words filled to width by
            #                   spaces.
            #      |A paragraph  that|
            #      |is     justified.|
            #
            # *Default*::       <tt>Text::Format::LEFT_ALIGN</tt>
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        def format_style=(fs)
            if fs < LEFT_ALIGN || fs > JUSTIFY
                raise ArgumentError,
                      "Invalid value provided for format_style."
            end
            @format_style = fs
        end

            # Indicates that the format style is left alignment.
            #
            # *Default*::       +true+
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        def left_align?
            return @format_style == LEFT_ALIGN
        end

            # Indicates that the format style is right alignment.
            #
            # *Default*::       +false+
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        def right_align?
            return @format_style == RIGHT_ALIGN
        end

            # Indicates that the format style is right fill.
            #
            # *Default*::       +false+
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        def right_fill?
            return @format_style == RIGHT_FILL
        end

            # Indicates that the format style is full justification.
            #
            # *Default*::       +false+
            # <b>Used in</b>::  <tt>#format</tt>, <tt>#paragraphs</tt>
        def justify?
            return @format_style == JUSTIFY
        end

    private
        def __format(to_wrap) #:nodoc:
            findent_str = ' ' * @first_indent
            bindent_str = ' ' * @body_indent

            words = to_wrap.split(/\s+/).compact
            words.shift if words[0].empty?
            to_wrap = []

            abbrev = false
            width = @columns - @first_indent - @left_margin - @right_margin
            line = words.shift
            abbrev = __is_abbrev(line) unless line.nil? || line.empty?

            w = words.shift
            while not w.nil?
                if (w.size + line.size < (width - 1)) ||
                   (line !~ LEQ_RE || abbrev) &&
                   (w.size + line.size < width)
                    line << " " if (line =~ LEQ_RE) && (!abbrev)
                    line << " #{w}"
                else
                    break
                end

                abbrev = __is_abbrev(w)
                w = words.shift
            end

            line, w = __do_break(line, w) if @nobreak && (!line.nil?)
            to_wrap << __make_line(line, findent_str, width, w.nil?) unless
                line.nil?
            line = w
            width = @columns - @body_indent - @left_margin - @right_margin
            abbrev = false
            abbrev = __is_abbrev(line) unless line.nil?

            while w = words.shift
                if (w.size + line.size < (width - 1)) ||
                   (line !~ LEQ_RE || abbrev) &&
                   (w.size + line.size < width)
                    line << " " if (line =~ LEQ_RE) && (!abbrev)
                    line << " #{w}"
                else
                    line, w = __do_break(line, w) if @nobreak
                    to_wrap << __make_line(line, bindent_str, width, w.nil?) unless
                        line.nil?
                    line = w
                end

                abbrev = __is_abbrev(w) unless w.nil?
            end

            unless line.nil?
                to_wrap << __make_line(line, bindent_str, width, true)
            end

            if @tag_paragraph && to_wrap.size > 0
                clr = %r{`(\w+)'}.match([caller(1)].flatten[0])[1]
                clr = "" if clr.nil?
                @tag_cur = @tag_text[0] if (not @tag_text[0].nil?) &&
                    (@tag_cur.size < 1) && (clr != "__paragraphs")
                fchar = /(\S)/.match(to_wrap[0])[1]
                white = to_wrap[0].index(fchar)
                if ((white - @left_margin - 1) > @tag_cur.size)
                    white = @tag_cur.size + @left_margin
                    to_wrap[0].gsub!(/^ {#{white}}/,
                                    "#{' ' * @left_margin}#{@tag_cur}")
                else
                    to_wrap.unshift("#{' ' * @left_margin}#{@tag_cur}\n")
                end
            end
            to_wrap.join('')
        end

            # format lines in text into paragraphs with each element of @wrap
            # a paragraph; uses Text::Format.format for the formatting
        def __paragraphs(to_wrap) #:nodoc:
            if (@first_indent == @body_indent) || @tag_paragraph
                p_end = "\n"
            else
                p_end = ''
            end

            cnt = 0
            ret = []
            to_wrap.each do |tw|
                @tag_cur = @tag_text[cnt] if @tag_paragraph
                @tag_cur = '' if @tag_cur.nil?
                line = __format(tw)
                ret << "#{line}#{p_end}" if (not line.nil?) && line.size > 0
                cnt += 1
            end

            ret[-1].chomp!
            ret.join('')
        end

            # center text using spaces on left side to pad it out empty lines
            # are preserved
        def __center(to_center) #:nodoc:
            tabs = 0
            width = @columns - @left_margin - @right_margin
            centered = []
            to_center.each do |tc|
                s = tc.strip
                tabs = s.count("\t")
                tabs = 0 if tabs.nil?
                ct = ((width - s.size - (tabs * @tabstop) + tabs) / 2)
                ct = (width - @left_margin - @right_margin) - ct
                centered << "#{s.rjust(ct)}\n"
            end
            centered.join('')
        end

            # expand tabs to spaces
            # should be similar to Text::Tabs::expand
        def __expand(to_expand) #:nodoc:
            expanded = []
            to_expand.split("\n").each do |te|
                expanded << te.gsub(/\t/, ' ' * @tabstop)
            end
            expanded.join('')
        end

        def __unexpand(to_unexpand) #:nodoc:
            unexpanded = []
            to_unexpand.split("\n").each do |tu|
                unexpanded << tu.gsub(/ {#{@tabstop}}/, "\t")
            end
            unexpanded.join('')
        end

        def __is_abbrev(word) #:nodoc:
                # remove period if there is one.
            w = word.gsub(/\.$/, '') unless word.nil?
            if (!@extra_space || ABBREV.include?(w) ||
                @abbreviations.include?(w))
                true
            else
                false
            end
        end

        def __make_line(line, indent, width, last = false) #:nodoc:
            fill = ""
            lmargin = " " * @left_margin
            fill = " " * (width - line.size) if right_fill?

            if (justify? && (!line.nil? && !line.empty?) &&
                line =~ /\S+\s+\S+/ && !last)
                spaces = width - line.size
                words = line.split(/(\s+)/)
                ws = spaces / (words.size / 2)
                spaces = spaces % (words.size / 2) if ws > 0
                words.reverse.each do |rw|
                    next if rw =~ /^\S/
                    rw.sub!(/^/, " " * ws)
                    (spaces > 0) or next
                    rw.sub!(/^/, " ")
                    spaces -= 1
                end
                line = words.join('')
            end
            line = "#{lmargin}#{indent}#{line}#{fill}\n" unless line.nil?
            if right_align? && !line.nil?
                line.sub(/^/, " " * (@columns - @right_margin - (line.size -
                                                                 1)))
            else
                line
            end
        end

        def __do_break(line, next_line) #:nodoc:
            no_brk = false
            words = []
            words = line.split(/\s+/) unless line.nil?
            last_word = words[-1]

            @nobreak_regex.each do |k, v|
                no_brk = true if last_word =~ /#{k}/ &&
                    next_line =~ /#{v}/
            end

            if no_brk && words.size > 1
                i = words.size
                while i > 0
                    no_brk = false
                    @nobreak_regex.each do |k, v|
                        no_brk = true if words[i + 1] =~ /#{k}/ &&
                            words[i] =~ /#{v}/
                    end
                    i -= 1
                    break if not no_brk
                end
                if i > 0
                    (l, m1, m2) = brk_re(i).match(line)
                    line.sub!(brk_re(i), m1)
                    next_line = "#{m2} #{next_line}"
                    line.sub!(/\s+$/, '')
                end
            end
            [line, next_line]
        end

        def __create(arg = nil, &block) #:nodoc:
                # Format::Text.new(text-to-wrap)
            @text = arg unless arg.nil?
                # Defaults
            @columns        = 72
            @tabstop        = 8
            @first_indent   = 4
            @body_indent    = 0
            @format_style   = LEFT_ALIGN
            @left_margin    = 0
            @right_margin   = 0
            @extra_space    = false
            @text           = Array.new if @text.nil?
            @tag_paragraph  = false
            @tag_text       = Array.new
            @tag_cur        = ""
            @abbreviations  = Array.new
            @nobreak        = false
            @nobreak_regex  = Hash.new

            instance_eval(&block) unless block.nil?
        end

    public
            # Formats text into a nice paragraph format. The text is separated
            # into words and then reassembled a word at a time using the
            # settings of this Format object. If a word is larger than the
            # number of columns available for formatting, then word will
            # appear on the line by itself.
            #
            # If +to_wrap+ is +nil+, then the value of <tt>#text</tt> will be
            # worked on.
        def format(to_wrap = nil)
            to_wrap = @text if to_wrap.nil?
            if to_wrap.class == Array
                __format(to_wrap[0])
            else
                __format(to_wrap)
            end
        end

            # Considers each element of text (provided or internal) as a
            # paragraph. If <tt>#first_indent</tt> is the same as
            # <tt>#body_indent</tt>, then paragraphs will be separated by a
            # single empty line in the result; otherwise, the paragraphs will
            # follow immediately after each other. Uses <tt>#format</tt> to do
            # the heavy lifting.
        def paragraphs(to_wrap = nil)
            to_wrap = @text if to_wrap.nil?
            __paragraphs([to_wrap].flatten)
        end

            # Centers the text, preserving empty lines and tabs.
        def center(to_center = nil)
            to_center = @text if to_center.nil?
            __center([to_center].flatten)
        end

            # Replaces all tab characters in the text with <tt>#tabstop</tt>
            # spaces.
        def expand(to_expand = nil)
            to_expand = @text if to_expand.nil?
            if to_expand.class == Array
                to_expand.collect { |te| __expand(te) }
            else
                __expand(to_expand)
            end
        end

            # Replaces all occurrences of <tt>#tabstop</tt> consecutive spaces
            # with a tab character.
        def unexpand(to_unexpand = nil)
            to_unexpand = @text if to_unexpand.nil?
            if to_unexpand.class == Array
                to_unexpand.collect { |te| v << __unexpand(te) }
            else
                __unexpand(to_unexpand)
            end
        end

            # This constructor takes advantage of a technique for Ruby object
            # construction introduced by Andy Hunt and Dave Thomas (see
            # reference), where optional values are set using commands in a
            # block.
            #
            #   Text::Format.new {
            #       columns         = 72
            #       left_margin     = 0
            #       right_margin    = 0
            #       first_indent    = 4
            #       body_indent     = 0
            #       format_style    = Text::Format::LEFT_ALIGN
            #       extra_space     = false
            #       abbreviations   = {}
            #       tag_paragraph   = false
            #       tag_text        = []
            #       nobreak         = false
            #       nobreak_regex   = {}
            #       tabstop         = 8
            #       text            = nil
            #   }
            #
            # As shown above, +arg+ is optional. If +arg+ is specified and is
            # a +String+, then arg is used as the default value of
            # <tt>#text</tt>. Alternately, an existing Text::Format object can
            # be used or a Hash can be used. With all forms, a block can be
            # specified.
            #
            # *Reference*:: "Object Construction and Blocks"
            #               <http://www.pragmaticprogrammer.com/ruby/articles/insteval.html>
            #
        def initialize(arg = nil, &block)
            case arg
                when Text::Format
                    __create(arg.text) do
                        @columns        = arg.columns
                        @tabstop        = arg.tabstop
                        @first_indent   = arg.first_indent
                        @body_indent    = arg.body_indent
                        @format_style   = arg.format_style
                        @left_margin    = arg.left_margin
                        @right_margin   = arg.right_margin
                        @extra_space    = arg.extra_space
                        @tag_paragraph  = arg.tag_paragraph
                        @tag_text       = arg.tag_text
                        @abbreviations  = arg.abbreviations
                        @nobreak        = arg.nobreak
                        @nobreak_regex  = arg.nobreak_regex
                        @text           = arg.text
                    end
                    instance_eval(&block) unless block.nil?
                when Hash
                    __create do
                        @columns        = arg['columns'] if arg.has_key?('columns')
                        @tabstop         = arg['tabstop'] if arg.has_key?('tabstop')
                        @first_indent    = arg['first_indent'] if arg.has_key?('first_indent')
                        @body_indent     = arg['body_indent'] if arg.has_key?('body_indent')
                        @format_style    = arg['format_style'] if arg.has_key?('format_style')
                        @left_margin     = arg['left_margin'] if arg.has_key?('left_margin')
                        @right_margin    = arg['right_margin'] if arg.has_key?('right_margin')
                        @extra_space     = arg['extra_space'] if arg.has_key?('extra_space')
                        @text            = arg['text'] if arg.has_key?('text')
                        @tag_paragraph   = arg['tag_paragraph'] if arg.has_key?('tag_paragraph')
                        @tag_text        = arg['tag_text'] if arg.has_key?('tag_text')
                        @abbreviations   = arg['abbreviations'] if arg.has_key?('abbreviations')
                        @nobreak         = arg['nobreak'] if arg.has_key?('nobreak')
                        @nobreak_regex   = arg['nobreak_regex'] if arg.has_key?('nobreak_regex')
                    end
                    instance_eval(&block) unless block.nil?
                when String
                    __create(arg, &block)
                when NilClass
                    __create(&block)
                else
                    raise TypeError
            end
        end
    end
end

if __FILE__ == $0
    require 'test/unit'

    class TestTEXT__Format < Test::Unit::TestCase #:nodoc:
        attr_accessor :format_o

        GETTYSBURG = <<'EOS'
    Four score and seven years ago our fathers brought forth on this
    continent a new nation, conceived in liberty and dedicated to the
    proposition that all men are created equal. Now we are engaged in
    a great civil war, testing whether that nation or any nation so
    conceived and so dedicated can long endure. We are met on a great
    battlefield of that war. We have come to dedicate a portion of
    that field as a final resting-place for those who here gave their
    lives that that nation might live. It is altogether fitting and
    proper that we should do this. But in a larger sense, we cannot
    dedicate, we cannot consecrate, we cannot hallow this ground.
    The brave men, living and dead who struggled here have consecrated
    it far above our poor power to add or detract. The world will
    little note nor long remember what we say here, but it can never
    forget what they did here. It is for us the living rather to be
    dedicated here to the unfinished work which they who fought here
    have thus far so nobly advanced. It is rather for us to be here
    dedicated to the great task remaining before us--that from these
    honored dead we take increased devotion to that cause for which
    they gave the last full measure of devotion--that we here highly
    resolve that these dead shall not have died in vain, that this
    nation under God shall have a new birth of freedom, and that
    government of the people, by the people, for the people shall
    not perish from the earth.

            -- Pres. Abraham Lincoln, 19 November 1863
EOS

            # Tests both abbreviations and abbreviations=
        def test_abbreviations
            abbr = ["    Pres. Abraham Lincoln\n",
                    "    Pres.  Abraham Lincoln\n"]
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal([], @format_o.abbreviations)
            assert_nothing_raised {
                @format_o.abbreviations = [ 'foo', 'bar' ]
            }
            assert_equal([ 'foo', 'bar' ], @format_o.abbreviations)
            assert_equal(abbr[0], @format_o.format(abbr[0]))
            assert_nothing_raised { @format_o.extra_space = true }
            assert_equal(abbr[1], @format_o.format(abbr[0]))
            assert_nothing_raised { @format_o.abbreviations = [ "Pres" ] }
            assert_equal([ "Pres" ], @format_o.abbreviations)
            assert_equal(abbr[0], @format_o.format(abbr[0]))
            assert_nothing_raised { @format_o.extra_space = false }
            assert_equal(abbr[0], @format_o.format(abbr[0]))
        end

            # Tests both body_indent and body_indent=
        def test_body_indent
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal(0, @format_o.body_indent)
            assert_nothing_raised { @format_o.body_indent = 7 }
            assert_equal(7, @format_o.body_indent)
            assert_nothing_raised { @format_o.body_indent = -3 }
            assert_equal(3, @format_o.body_indent)
            assert_nothing_raised { @format_o.body_indent = "9" }
            assert_equal(9, @format_o.body_indent)
            assert_nothing_raised { @format_o.body_indent = "-2" }
            assert_equal(2, @format_o.body_indent)
            assert_match(/^  [^ ]/,
                         @format_o.format(GETTYSBURG).split("\n")[1])
        end

            # Tests both columns and columns=
        def test_columns
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal(72, @format_o.columns)
            assert_nothing_raised { @format_o.columns = 7 }
            assert_equal(7, @format_o.columns)
            assert_nothing_raised { @format_o.columns = -3 }
            assert_equal(3, @format_o.columns)
            assert_nothing_raised { @format_o.columns = "9" }
            assert_equal(9, @format_o.columns)
            assert_nothing_raised { @format_o.columns = "-2" }
            assert_equal(2, @format_o.columns)
            assert_nothing_raised { @format_o.columns = 40 }
            assert_equal(40, @format_o.columns)
            assert_match(/this continent$/,
                         @format_o.format(GETTYSBURG).split("\n")[1])
        end

            # Tests both extra_space and extra_space=
        def test_extra_space
            assert_nothing_raised { @format_o = Text::Format.new }
            assert(!@format_o.extra_space)
            assert_nothing_raised { @format_o.extra_space = true }
            assert(@format_o.extra_space)
                # The behaviour of extra_space is tested in
                # test_abbreviations. There is no need to reproduce it here.
        end

            # Tests both first_indent and first_indent=
        def test_first_indent
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal(4, @format_o.first_indent)
            assert_nothing_raised { @format_o.first_indent = 7 }
            assert_equal(7, @format_o.first_indent)
            assert_nothing_raised { @format_o.first_indent = -3 }
            assert_equal(3, @format_o.first_indent)
            assert_nothing_raised { @format_o.first_indent = "9" }
            assert_equal(9, @format_o.first_indent)
            assert_nothing_raised { @format_o.first_indent = "-2" }
            assert_equal(2, @format_o.first_indent)
            assert_match(/^  [^ ]/,
                         @format_o.format(GETTYSBURG).split("\n")[0])
        end

        def test_format_style
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal(Text::Format::LEFT_ALIGN, @format_o.format_style)
            assert_match(/^November 1863$/,
                         @format_o.format(GETTYSBURG).split("\n")[-1])
            assert_nothing_raised {
                @format_o.format_style = Text::Format::RIGHT_ALIGN
            }
            assert_equal(Text::Format::RIGHT_ALIGN, @format_o.format_style)
            assert_match(/^ +November 1863$/,
                         @format_o.format(GETTYSBURG).split("\n")[-1])
            assert_nothing_raised { @format_o.format_style = Text::Format::RIGHT_FILL }
            assert_equal(Text::Format::RIGHT_FILL, @format_o.format_style)
            assert_match(/^November 1863 +$/,
                         @format_o.format(GETTYSBURG).split("\n")[-1])
            assert_nothing_raised { @format_o.format_style = Text::Format::JUSTIFY }
            assert_equal(Text::Format::JUSTIFY, @format_o.format_style)
            assert_match(/^of freedom, and that government of the people, by the  people,  for  the$/,
                         @format_o.format(GETTYSBURG).split("\n")[-3])
            assert_raises(ArgumentError) { @format_o.format_style = 33 }
        end

        def test_tag_paragraph
            assert_nothing_raised { @format_o = Text::Format.new }
            assert(!@format_o.tag_paragraph)
            assert_nothing_raised { @format_o.tag_paragraph = true }
            assert(@format_o.tag_paragraph)
            assert_not_equal(@format_o.paragraphs([GETTYSBURG, GETTYSBURG]),
                             Text::Format.new.paragraphs([GETTYSBURG,
                                                          GETTYSBURG]))
        end

        def test_tag_text
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal([], @format_o.tag_text)
            assert_equal(@format_o.format(GETTYSBURG),
                         Text::Format.new.format(GETTYSBURG))
            assert_nothing_raised {
                @format_o.tag_paragraph = true
                @format_o.tag_text = ["Gettysburg Address", "---"]
            }
            assert_not_equal(@format_o.format(GETTYSBURG),
                             Text::Format.new.format(GETTYSBURG))
            assert_not_equal(@format_o.paragraphs([GETTYSBURG, GETTYSBURG]),
                             Text::Format.new.paragraphs([GETTYSBURG,
                                                          GETTYSBURG]))
            assert_not_equal(@format_o.paragraphs([GETTYSBURG, GETTYSBURG,
                                                   GETTYSBURG]),
                             Text::Format.new.paragraphs([GETTYSBURG,
                                                          GETTYSBURG,
                                                          GETTYSBURG]))
        end

        def test_justify?
            assert_nothing_raised { @format_o = Text::Format.new }
            assert(!@format_o.justify?)
            assert_nothing_raised {
                @format_o.format_style = Text::Format::RIGHT_ALIGN
            }
            assert(!@format_o.justify?)
            assert_nothing_raised {
                @format_o.format_style = Text::Format::RIGHT_FILL
            }
            assert(!@format_o.justify?)
            assert_nothing_raised {
                @format_o.format_style = Text::Format::JUSTIFY
            }
            assert(@format_o.justify?)
                # The format testing is done in test_format_style
        end

        def test_left_align?
            assert_nothing_raised { @format_o = Text::Format.new }
            assert(@format_o.left_align?)
            assert_nothing_raised { @format_o.format_style = Text::Format::RIGHT_ALIGN }
            assert(!@format_o.left_align?)
            assert_nothing_raised { @format_o.format_style = Text::Format::RIGHT_FILL }
            assert(!@format_o.left_align?)
            assert_nothing_raised { @format_o.format_style = Text::Format::JUSTIFY }
            assert(!@format_o.left_align?)
                # The format testing is done in test_format_style
        end

        def test_left_margin
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal(0, @format_o.left_margin)
            assert_nothing_raised { @format_o.left_margin = -3 }
            assert_equal(3, @format_o.left_margin)
            assert_nothing_raised { @format_o.left_margin = "9" }
            assert_equal(9, @format_o.left_margin)
            assert_nothing_raised { @format_o.left_margin = "-2" }
            assert_equal(2, @format_o.left_margin)
            assert_nothing_raised { @format_o.left_margin = 7 }
            assert_equal(7, @format_o.left_margin)
            assert_nothing_raised {
                ft = @format_o.format(GETTYSBURG).split("\n")
                assert_match(/^ {11}Four score/, ft[0])
                assert_match(/^ {7}November/, ft[-1])
            }
        end

            # Tests both nobreak and nobreak_regex, since one is only useful
            # with the other.
        def test_nobreak
            assert_nothing_raised { @format_o = Text::Format.new }
            assert(!@format_o.nobreak)
            assert(@format_o.nobreak_regex.empty?)
            assert_nothing_raised {
                @format_o.nobreak = true
                @format_o.nobreak_regex = { '^this$' => '^continent$' }
                @format_o.columns = 77
            }
            assert(@format_o.nobreak)
            assert_equal({ '^this$' => '^continent$' }, @format_o.nobreak_regex)
            assert_match(/^this continent/,
                         @format_o.format(GETTYSBURG).split("\n")[1])
        end

        def test_right_align?
            assert_nothing_raised { @format_o = Text::Format.new }
            assert(!@format_o.right_align?)
            assert_nothing_raised { @format_o.format_style = Text::Format::RIGHT_ALIGN }
            assert(@format_o.right_align?)
            assert_nothing_raised { @format_o.format_style = Text::Format::RIGHT_FILL }
            assert(!@format_o.right_align?)
            assert_nothing_raised { @format_o.format_style = Text::Format::JUSTIFY }
            assert(!@format_o.right_align?)
                # The format testing is done in test_format_style
        end

        def test_right_fill?
            assert_nothing_raised { @format_o = Text::Format.new }
            assert(!@format_o.right_fill?)
            assert_nothing_raised {
                @format_o.format_style = Text::Format::RIGHT_ALIGN
            }
            assert(!@format_o.right_fill?)
            assert_nothing_raised {
                @format_o.format_style = Text::Format::RIGHT_FILL
            }
            assert(@format_o.right_fill?)
            assert_nothing_raised {
                @format_o.format_style = Text::Format::JUSTIFY
            }
            assert(!@format_o.right_fill?)
                # The format testing is done in test_format_style
        end

        def test_right_margin
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal(0, @format_o.right_margin)
            assert_nothing_raised { @format_o.right_margin = -3 }
            assert_equal(3, @format_o.right_margin)
            assert_nothing_raised { @format_o.right_margin = "9" }
            assert_equal(9, @format_o.right_margin)
            assert_nothing_raised { @format_o.right_margin = "-2" }
            assert_equal(2, @format_o.right_margin)
            assert_nothing_raised { @format_o.right_margin = 7 }
            assert_equal(7, @format_o.right_margin)
            assert_nothing_raised {
                ft = @format_o.format(GETTYSBURG).split("\n")
                assert_match(/^ {4}Four score.*forth on$/, ft[0])
                assert_match(/^November/, ft[-1])
            }
        end

        def test_tabstop
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal(8, @format_o.tabstop)
            assert_nothing_raised { @format_o.tabstop = 7 }
            assert_equal(7, @format_o.tabstop)
            assert_nothing_raised { @format_o.tabstop = -3 }
            assert_equal(3, @format_o.tabstop)
            assert_nothing_raised { @format_o.tabstop = "9" }
            assert_equal(9, @format_o.tabstop)
            assert_nothing_raised { @format_o.tabstop = "-2" }
            assert_equal(2, @format_o.tabstop)
        end

        def test_text
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal([], @format_o.text)
            assert_nothing_raised { @format_o.text = "Test Text" }
            assert_equal("Test Text", @format_o.text)
            assert_nothing_raised { @format_o.text = ["Line 1", "Line 2"] }
            assert_equal(["Line 1", "Line 2"], @format_o.text)
        end

        def test_s_new
                # new(NilClass) { block }
            assert_nothing_raised do
                @format_o = Text::Format.new {
                    self.text = "Test 1, 2, 3"
                }
            end
            assert_equal("Test 1, 2, 3", @format_o.text)
                # new(Hash)
            assert_nothing_raised do
                @format_o = Text::Format.new('columns' => 72)
            end
            assert_equal(72, @format_o.columns)
                # new(Hash) { block }
            assert_nothing_raised do
                @format_o = Text::Format.new('columns' => 80) {
                    self.text = "Test 4, 5, 6"
                }
            end
            assert_equal("Test 4, 5, 6", @format_o.text)
            assert_equal(80, @format_o.columns)
                # new(Text::Format)
            assert_nothing_raised do
                fo = Text::Format.new(@format_o)
                assert(fo == @format_o)
            end
                # new(Text::Format) { block }
            assert_nothing_raised do
                fo = Text::Format.new(@format_o) {
                    self.columns = 79
                }
                assert(fo != @format_o)
            end
                # new(String)
            assert_nothing_raised do
                @format_o = Text::Format.new("Test A, B, C")
            end
            assert_equal("Test A, B, C", @format_o.text)
                # new(String) { block }
            assert_nothing_raised do
                @format_o = Text::Format.new("Test X, Y, Z") {
                    self.columns = -5
                }
            end
            assert_equal("Test X, Y, Z", @format_o.text)
            assert_equal(5, @format_o.columns)
        end

        def test_center
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_nothing_raised {
                ct = @format_o.center(GETTYSBURG.split("\n")).split("\n")
                assert_match(/^    Four score and seven years ago our fathers brought forth on this/,
                             ct[0])
                assert_match(/^                       not perish from the earth./,
                             ct[-3])
            }
        end

        def test_expand
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal("          ", @format_o.expand("\t  "))
            assert_nothing_raised { @format_o.tabstop = 4 }
            assert_equal("      ", @format_o.expand("\t  "))
        end

        def test_unexpand
            assert_nothing_raised { @format_o = Text::Format.new }
            assert_equal("\t  ", @format_o.unexpand("          "))
            assert_nothing_raised { @format_o.tabstop = 4 }
            assert_equal("\t  ", @format_o.unexpand("      "))
        end
    end
end
